#include <sys/types.h>
#include <sys/stat.h>

#include <malloc.h>
#include <fcntl.h>
#include <stdint.h>
#include <unistd.h>

#include <ft2build.h>
#include FT_FREETYPE_H
#include "config.h"
#include "libpdf/pdffont.h"
#include "libpdf/pdfliteral.h"

#ifndef HAVE_FT_GET_FIRST
#define FT_Get_First_Char(face, agindex) 
#define FT_Get_Next_Char(face, char_code, agindex) (char_code+1)
#endif

/* what I want from standard fonts...
   I want to specify the normal, external font name, and have (1) a
   new font object defined, (2) the font object inserted into the font
   subdictionary of the document's resources dictionary, (3) a name
   generated for the font, and (4) that name returned.

   That means a font should be typedefed to a name, and the _new
   procedure should also include the document.

   So what I'll do is to look up the Resources dictionary, and then
   the Fonts subdictionary.  I'll look at the last font in the
   subdictionary, parse it to get the font number, and generate a new
   font number.  Then I'll define the dictionary defining the new
   font, and insert it in the font subdictionary (as an indirect
   object).  Then when I want to use a font with a text string, I can
   insert the name returned from here, and a size. */

/* lots of shared code between these functions...  need to merge */
PDF_Font PDF_Font_new_standard(PDF_Document doc, PDF_FontName name)
{
	/* NOTE:  this array MUST correspond to the definition of
	   PDF_FontName in pdffont.h */
	static PDF_CString fontname[] = {
	  "/Courier",
	  "/Courier-Bold",
	  "/Courier-Oblique",
	  "/Courier-BoldOblique",
	  "/Helvetica",
	  "/Helvetica-Bold",
	  "/Helvetica-Oblique",
	  "/Helvetica-BoldOblique",
	  "/Times-Roman",
	  "/Times-Bold",
	  "/Times-Italic",
	  "/Times-BoldItalic",
	  "/Symbol",
	  "/ZapfDingbats"
	};

	PDF_Dictionary fontdef, pages, resources, fonts;
	
	fontdef = NULL;

	if ((name < 0) || (name > 13))
		return NULL;
	
	pages = (PDF_Dictionary) PDF_Dictionary_get_entry(doc, "/Pages");
	resources = (PDF_Dictionary) PDF_Dictionary_get_entry(pages, "/Resources");
	fonts = (PDF_Dictionary) PDF_Dictionary_get_entry(resources, "/Font");
	
	if (((fontdef = PDF_Dictionary_new()) == NULL) ||
		
		(PDF_Dictionary_add_entry(fontdef, "/Type", (PDF_Object) PDF_Literal_new("/Font")) == PDF_CBOOL_FALSE) ||
		(PDF_Dictionary_add_entry(fontdef, "/Subtype", (PDF_Object) PDF_Literal_new("/Type1")) == PDF_CBOOL_FALSE) ||
		(PDF_Dictionary_add_entry(fontdef, "/BaseFont", (PDF_Object) PDF_Literal_new(fontname[name])) == PDF_CBOOL_FALSE)) {
		
		if (fontdef != NULL) free(fontdef);
		return NULL;
	}

	PDF_Object_set_indirect((PDF_Object) fontdef, PDF_CBOOL_TRUE);
	
	PDF_Dictionary_add_entry(fonts, fontname[name], (PDF_Object) fontdef);
	
	return fontdef;
}

PDF_Font PDF_Font_new_embed(PDF_Document doc, PDF_CString fontdata, int fontsize)
{
	FT_Library ftlib;
	FT_Error error;
	FT_Face ftface;
	FT_ULong firstchar, lastchar, undefwid;
	FT_ULong ftchar, ftoldchar, fttemp;
	FT_UInt  ftidx;
	/* for some reason, using PDF_CString in the next line has it
	   thinking the pointer, not the buffer contents, is the const */
	const char* fontname;
	char fontnamebuf[PDF_SMALLBUFSIZ];
	int success;
	int lastwid;
	
	PDF_Dictionary fontdef, fontdesc;
	PDF_Stream fontstream;
	PDF_Dictionary pages, resources, fonts;
	PDF_Array width_array, bbox;
	PDF_Font_Flags flags;

	/* Initialize the Freetype library */
	error = FT_Init_FreeType(&ftlib);
	if (error != 0)
		return NULL;

	/* Get the font face from the in-memory truetype file */
	error = FT_New_Memory_Face(ftlib, fontdata, fontsize, 0, &ftface);
	if (error != 0)
		return NULL;

	/* start filling dictionaries... */
	fontdef = PDF_Dictionary_new();
	PDF_Object_set_indirect((PDF_Object) fontdef, PDF_CBOOL_TRUE);
	
	PDF_Dictionary_add_entry(fontdef, "/Type", (PDF_Object) PDF_Literal_new("/Font"));
	PDF_Dictionary_add_entry(fontdef, "/Subtype", (PDF_Object) PDF_Literal_new("/TrueType"));
							 
	/* Get the PostScript name */
	fontname = FT_Get_Postscript_Name(ftface);
	if (fontname == NULL) {
		FT_Done_FreeType(ftlib);
		return NULL;
	}
	snprintf(fontnamebuf, PDF_SMALLBUFSIZ, "/%s", fontname);
	PDF_Dictionary_add_entry(fontdef, "/BaseFont", (PDF_Object) PDF_Literal_new(fontnamebuf));

	/* Set the font flags.  We can get some from the font face header;
	   other have to come from inspecting the glyphs in the font.  We
	   start by assuming all caps; we may change this as we go through
	   the glyphs */
	flags = PDF_FONT_ALLCAP;

	/* Now the flags that come from the header */
	if (FT_IS_FIXED_WIDTH(ftface))
		flags |= PDF_FONT_FIXEDPITCH;

	if (ftface->style_flags & FT_STYLE_FLAG_ITALIC)
		flags |= PDF_FONT_ITALIC;

	if (ftface->style_flags & FT_STYLE_FLAG_BOLD)
	flags |= PDF_FONT_FORCEBOLD;
	
	/* Now have to cycle through getting the widths of all the
	   characters, and checking whether there are any non-upper case
	   or symbols.  A real nasty one here is I need the width of
	   undefined chars; I don't see a good way to get freetype2 to
	   give this to me.  So I need to try to find an undefined char,
	   load it, and find out how wide it is.  My heuristics will be
	   (1) if the first defined char is > 0 try one less than it; (2)
	   if there's a gap in the encoding get the char from the gap; (3)
	   get one character past the highest defined.  So far it seems
	   like the first heuristic is getting it right.  */
	width_array = PDF_Array_new();

	/* Get the first character */
	undefwid = 0;
	firstchar = FT_Get_First_Char(ftface, &ftidx);
	if (firstchar > 0) {
		error = FT_Load_Char(ftface, ftchar, FT_LOAD_NO_SCALE);
		undefwid = ftface->glyph->metrics.horiAdvance;
	}
	ftchar = firstchar;
	PDF_Dictionary_add_entry(fontdef, "/FirstChar", (PDF_Object) PDF_Int_new(ftchar));
	ftoldchar = ftchar;

	while (ftidx != 0) {
		// fprintf(stderr, "%d\n", ftchar);
 		if ((undefwid == 0) && (ftchar > ftoldchar+1)) {
			error = FT_Load_Char(ftface, ftchar, FT_LOAD_NO_SCALE);
			undefwid = ftface->glyph->metrics.horiAdvance;
		}
			
		for (fttemp = ftoldchar + 1; fttemp < ftchar; fttemp++)
			PDF_Queue_append(width_array, (PDF_Object) PDF_Int_new(0));

		/* Sigh.  What possible difference can it make to PDF if the
		   character set has characters outside Adobe's ``standard''
		   Latin-1, or lower-case letters?  Regardless, flags need to
		   be set...   This definition of the ``noncharacter'' chars
		   comes from inspection of Appendix D.1 of the PDF manual */
		if (((ftchar < 0040) || (ftchar > 0176)) &&
			((ftchar < 0241) || (ftchar > 0257)) &&
			((ftchar < 0261) || (ftchar > 0264)) &&
			((ftchar < 0266) || (ftchar > 0275)) &&
			((ftchar < 0277) || (ftchar > 0277)) &&
			((ftchar < 0301) || (ftchar > 0310)) &&
			((ftchar < 0312) || (ftchar > 0313)) &&
			((ftchar < 0315) || (ftchar > 0320)) &&
			((ftchar < 0341) || (ftchar > 0341)) &&
			((ftchar < 0343) || (ftchar > 0343)) &&
			((ftchar < 0350) || (ftchar > 0353)) &&
			((ftchar < 0361) || (ftchar > 0361)) &&
			((ftchar < 0365) || (ftchar > 0365)) &&
			((ftchar < 0370) || (ftchar > 0373)))
			flags |= PDF_FONT_SYMBOLIC;

		if (islower(ftchar))
			flags &= ~PDF_FONT_ALLCAP;
		
		error = FT_Load_Char(ftface, ftchar, FT_LOAD_NO_SCALE);

		/* PDF's definition of character width seems to correspond to
		   truetype's definition of horizontal advance, not width */
		PDF_Queue_append(width_array, (PDF_Object) PDF_Int_new(ftface->glyph->metrics.horiAdvance));
		
		ftoldchar = ftchar;
		ftchar = FT_Get_Next_Char(ftface, ftchar, &ftidx);
	}
	
	if (undefwid == 0) {
		error = FT_Load_Char(ftface, ftchar, FT_LOAD_NO_SCALE);
		undefwid = ftface->glyph->metrics.horiAdvance;
	}

	PDF_Dictionary_add_entry(fontdef, "/LastChar", (PDF_Object) PDF_Int_new(ftoldchar));
	PDF_Dictionary_add_entry(fontdef, "/Widths", (PDF_Object) width_array);

	/* If the symbolic flag isn't set, the nonsymbol needs to be.  PDF
	   Ref says this is for ``historical'' reasons */
	if (!(flags & PDF_FONT_SYMBOLIC))
		flags |= PDF_FONT_NONSYMBOLIC;
	
	fontdesc = PDF_Dictionary_new();
	PDF_Object_set_indirect((PDF_Object) fontdesc, PDF_CBOOL_TRUE);
	PDF_Dictionary_add_entry(fontdef, "/FontDescriptor", (PDF_Object) fontdesc);
	
	/* Font dictionary fields:
	   Type Font
	   Subtype TrueType
	   BaseFont (from TTF file)
	   FirstChar (need to get from TTF file somehow)
	   LastChar (ditto)
	   Widths (array -- ditto)
	   FontDescriptor (dictionary -- see later)
	 */

	PDF_Dictionary_add_entry(fontdesc, "/Type", (PDF_Object) PDF_Literal_new("/FontDescriptor"));
	PDF_Dictionary_add_entry(fontdesc, "/Ascent", (PDF_Object) PDF_Int_new(ftface->ascender));
	PDF_Dictionary_add_entry(fontdesc, "/Descent", (PDF_Object) PDF_Int_new(ftface->descender));
	PDF_Dictionary_add_entry(fontdesc, "/Flags", (PDF_Object) PDF_Int_new(flags));
	
	bbox = PDF_Array_new();
	PDF_Queue_append(bbox, (PDF_Object) PDF_Int_new(ftface->bbox.xMin));	
	PDF_Queue_append(bbox, (PDF_Object) PDF_Int_new(ftface->bbox.yMin));	
	PDF_Queue_append(bbox, (PDF_Object) PDF_Int_new(ftface->bbox.xMax));	
	PDF_Queue_append(bbox, (PDF_Object) PDF_Int_new(ftface->bbox.yMax));
	PDF_Dictionary_add_entry(fontdesc, "/FontBBox", (PDF_Object) bbox);

	PDF_Dictionary_add_entry(fontdesc, "/FontName", (PDF_Object) PDF_Literal_new(fontnamebuf));
	PDF_Dictionary_add_entry(fontdesc, "/MissingWidth", (PDF_Object) PDF_Int_new(undefwid));
	
	/* These two are supposed to be required, but I don't see how to
	   get them from freetype and both acroread and gv seem to render
	   just fine without them (not that I've tried any italics yet....) */
	/*	PDF_Dictionary_add_entry(fontdesc, "ItalicAngle", PDF_Int_new(0));
		PDF_Dictionary_add_entry(fontdesc, "StemV", PDF_Int_new(0));*/

	fontstream = PDF_Stream_new();
	PDF_Dictionary_add_entry(fontdesc, "/FontFile2", (PDF_Object) fontstream);
	
	/* Font descriptor fields:
	   Type FontDescriptor
	   Ascent (from TTF)
	   CapHeight (from TTF)
	   Descent (from TTF)
	   Flags (see below)
	   FontBBox (ttf)
	   FontName (same as BaseFont)
	   ItalicAngle (ttf?)
	   StemV (ttf?)
	   FontFile2 (actual ttf stream)
	*/

	PDF_Stream_set_param(fontstream, "/Length1", (PDF_Object) PDF_Int_new(fontsize));
	PDF_Queue_append(PDF_Stream_get_content(fontstream), (PDF_Object) PDF_Literal_connect(fontdata, fontsize));
	
	pages = (PDF_Dictionary) PDF_Dictionary_get_entry(doc, "/Pages");
	resources = (PDF_Dictionary) PDF_Dictionary_get_entry(pages, "/Resources");
	fonts = (PDF_Dictionary) PDF_Dictionary_get_entry(resources, "/Font");
	PDF_Dictionary_add_entry(fonts, fontnamebuf, (PDF_Object) fontdef);
	return fontdef;
}

PDF_Font PDF_Font_new_embed_file(PDF_Document doc, PDF_CString fname)
{
	int fontfd;
	PDF_CString fontdata;
	struct stat fontstat;
	PDF_Font retval;
	
	fontfd = open(fname, O_RDONLY);
	if (fontfd > 0) {
		fstat(fontfd, &fontstat);
		fontdata = malloc(fontstat.st_size);
		read(fontfd, fontdata, fontstat.st_size);
		retval = PDF_Font_new_embed(doc, fontdata, fontstat.st_size);

		close(fontfd);
	} else
		retval = NULL;

	return retval;
}

PDF_CString PDF_Font_get_basefont(PDF_Font obj)
{
	return PDF_Literal_get_value((PDF_Literal) PDF_Dictionary_get_entry((PDF_Dictionary) obj, "/BaseFont"));
}
