#include <assert.h>

#include "libpdf/pdfmisc.h"

#include "libpdf/pdffont.h"
#include "libpdf/pdfliteral.h"
#include "libpdf/pdfpage.h"
#include "libpdf/pdftext.h"

PDF_Text PDF_Text_new(PDF_Graphics graphics)
{
	PDF_Queue retval;

	retval = PDF_Queue_new("BT\n", "", "ET\n");
	if (graphics != NULL)
		PDF_Queue_append(graphics, (PDF_Object) retval);

	return (PDF_Text) retval;
}

/* Text state operators */
PDF_CBool PDF_Text_set_charSpace(PDF_Text text, float space)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%f Tc\n", space);
}

PDF_CBool PDF_Text_set_wordSpace(PDF_Text text, float space)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%f Tw\n", space);
}
	
PDF_CBool PDF_Text_set_scale(PDF_Text text, float scale)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%f Tz\n", scale);
}

PDF_CBool PDF_Text_set_leading(PDF_Text text, float leading)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%f TL\n", leading);
}

PDF_CBool PDF_Text_set_font(PDF_Text text, PDF_Font font, float size)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%s %f Tf\n", PDF_Font_get_basefont(font), size);
}

PDF_CBool PDF_Text_set_render(PDF_Text text, PDF_Text_Render_Mode rendermode)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%d Tr\n", rendermode);
}
	
PDF_CBool PDF_Text_set_rise(PDF_Text text, float rise)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%f Ts\n", rise);
}

/* Text positioning operators */
PDF_CBool PDF_Text_newline_xy(PDF_Text text, float xoff, float yoff)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%f %f Td\n", xoff, yoff);
}

PDF_CBool PDF_Text_newline_leading(PDF_Text text, float xoff, float yoff)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%f %f TD\n", xoff, yoff);
}

PDF_CBool PDF_Text_set_matrix(PDF_Text text, float a, float b, float c, float d, float e, float f)
{
	return PDF_Queue_append_op((PDF_Queue) text, "%f %f %f %f %f %f Tm\n", a, b, c, d, e, f);
}
	
PDF_CBool PDF_Text_newline(PDF_Text text)
{
	return PDF_Queue_append_op((PDF_Queue) text, "T*\n");
}

/* Text showing operators */
/* note: they all assume a short string. Fix on next go-round */

/* need to escape special characters in literal text strings.
   It's a bit kludgy to have this here, and not make use of the
   PDF_String_new function I've defined... */
static PDF_CString PDF_CString_escape_specials(PDF_CString string)
{
	PDF_CString retval;
	int size, i, j;
	
	// No way to do this in one pass.  First pass, find out how long
	// the buffer has to be.
	size = 1; // extra space for terminating null
	for (i = 0; string[i] != '\0'; i++)
		switch (string[i]) {
		  case '\n':
		  case '\r':
		  case '\t':
		  case '\b':
		  case '\f':
		  case '(':
		  case ')':
		  case '\\':
			size += 2;
			break;

		  default:
			size++;
		}

	retval = malloc(size);
	if (retval == NULL)
		return NULL;
	
	// there's gotta be a better way to do this!
	j = 0;
	for (i = 0; string[i] != '\0'; i++)
		switch (string[i]) {
		  case '\n':
			retval[j++] = '\\';
			retval[j++] = 'n';
			break;
			
		  case '\r':
			retval[j++] = '\\';
			retval[j++] = 'r';
			break;
			
		  case '\t':
			retval[j++] = '\\';
			retval[j++] = 't';
			break;
			
		  case '\b':
			retval[j++] = '\\';
			retval[j++] = 'b';
			break;
			
		  case '\f':
			retval[j++] = '\\';
			retval[j++] = 'f';
			break;
			
		  case '(':
			retval[j++] = '\\';
			retval[j++] = '(';
			break;
			
		  case ')':
			retval[j++] = '\\';
			retval[j++] = ')';
			break;
			
		  case '\\':
			retval[j++] = '\\';
			retval[j++] = '\\';
			break;

		  default:
			retval[j++] = string[i];
		}
	retval[j++] = '\0';
	
	assert(j == size);

	return retval;
}

PDF_CBool PDF_Text_string(PDF_Text text, PDF_CString string)
{
	
	PDF_CString temp = PDF_CString_escape_specials(string);
	PDF_CBool retval = PDF_Queue_append_op((PDF_Queue) text, "(%s) Tj\n", temp);
	free(temp);
	
	return retval;
}
	 
PDF_CBool PDF_Text_newline_string(PDF_Text text, PDF_CString string)
{
	PDF_CString temp = PDF_CString_escape_specials(string);
	PDF_CBool retval = PDF_Queue_append_op((PDF_Queue) text, "(%s) '\n", temp);
	free(temp);
	
	return retval;
}
	
	
PDF_CBool PDF_Text_space_string(PDF_Text text, float w, float c, PDF_CString string)
{
	PDF_CString temp = PDF_CString_escape_specials(string);
	PDF_CBool retval = PDF_Queue_append_op((PDF_Queue) text, "%d %d (%s) \"\n", w, c, temp);
	free(temp);
	
	return retval;
}

